/**
 *    This file is part of OtviAI.
 *
 *    OtviAI is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    OtviAI is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with OtviAI.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2008 Marnix Bakker, Willem de Neve, Michiel Konstapel and Otto Visser.
 * Suggestions/comments and bugs can go to: rondjeomdekerk@konstapel.nl
 */

require("RoadFinder/RoadFinder.nut");
/**
 * TODO:
 - do the whole thing in a transaction to try if you can do it for the money you have now?
 - check roads to depots; always generates err_unknown: aka I'm ignoring those also when I shouldn't
 - search with increasing radius: subtract inner circle to prevent that to be checked 10 times
 - max route cost based on max height found on map
 - add service to order?
 - replacing vehicles
 */

class OtviAI extends AIController {
	biggestTown = -1;	
	subsidyBuilt = false;
	
	myTownList = null;
	myStationsTileList = null;
	myLorryTileList = null;
	myDepotsTileList = null;
	
	passengerCargoId = -1;
	mailCargoId = -1;
	
	passengerEngine = -1;
	mailEngine = -1;
	
	lastMail1 = -1;
	lastMail2 = -1;
	lastMailVehicle = -1;
	lastTown1 = -1;
	lastTown2 = -1;
	lastPassengerVehicle = -1;
	lastMaxBusses = 0;
	expandMoneyMin = -1;
	
	pathfinder = null;
	MIN_DISTANCE = 50;	// ends at 40
	MAX_DISTANCE = 100;	// ends at 200

	uberstartTick = 0;

	function Save() {
		// TODO
		return {};
	}

	function Load() {
		Debug("TODO: implement load and save");
	}

	function Debug(s) {
		AILog.Info(GetDate() + ": " + s);
	}

	function Warning(s) {
		AILog.Warning(GetDate() + ": " + s);
	}

	function Error(s) {
		AILog.Error(GetDate() + ": " + s);
	}

	function GetDate() {
		local date = AIDate.GetCurrentDate();
		return "" + AIDate.GetYear(date) + "-" + AIDate.GetMonth(date) + "-" + AIDate.GetDayOfMonth(date);
	}

	function constructor() {
		pathfinder = RoadFinder();
		myStationsTileList = AIList();
		myLorryTileList = AIList();
		myDepotsTileList = AIList();
		myTownList = AITownList();

		pathfinder.cost.tile = 1;
		pathfinder.cost.max_cost = 30000;		// TODO: still an estimate
		pathfinder.cost.no_existing_road = 5;	// The cost that is added to _cost_tile if no road exists yet.
		pathfinder.cost.turn = 10;				// The cost that is added to _cost_tile if the direction changes.
		pathfinder.cost.slope =	15;				// The extra cost if a road tile is sloped.
		pathfinder.cost.bridge_per_tile = 20;	// The cost per tile of a new bridge, this is added to _cost_tile.
		pathfinder.cost.tunnel_per_tile = 20;	// The cost per tile of a new tunnel, this is added to _cost_tile.
		pathfinder.cost.coast =	20;				// The extra cost for a coast tile.
		pathfinder.cost.max_bridge_length = 10;	// The maximum length of a bridge that will be build.
		pathfinder.cost.max_tunnel_length = 10;	// The maximum length of a tunnel that will be build.
		
		local list = AICargoList();
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			if (AICargo.HasCargoClass(i, AICargo.CC_PASSENGERS) && AICargo.GetTownEffect(i) == AICargo.TE_PASSENGERS) {
				Debug("Found passenger cargo: " + AICargo.GetCargoLabel(i));
				passengerCargoId = i;
			}
		}

		list = AICargoList();
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			if (AICargo.HasCargoClass(i, AICargo.CC_MAIL) && AICargo.GetTownEffect(i) == AICargo.TE_MAIL) {
				Debug("Found mail cargo: " + AICargo.GetCargoLabel(i));
				mailCargoId = i;
			}
		}

		passengerEngine = getEngine(passengerCargoId);
		expandMoneyMin = 2 * AIEngine.GetPrice(passengerEngine);

		mailEngine = getEngine(mailCargoId);
		expandMoneyMin = (2 * AIEngine.GetPrice(mailEngine)) < expandMoneyMin ? expandMoneyMin : 2 * AIEngine.GetPrice(mailEngine);
		
		local towns = AITownList();
		towns.Valuate(AITown.GetPopulation);
		towns.Sort(AIAbstractList.SORT_BY_VALUE, false);
		biggestTown = towns.Begin();

		AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
	}

	function Start() {
		AICompany.SetName("OTVI");
		
		while(passengerEngine == -1 && mailEngine == -1) {
			manageLoan();	// no loan (also no hq) for as long as possible
			checkEvents();	// wait for suitable equipment
			
			this.Sleep(1);
		}

		buildHQ(biggestTown);
		
		// the main control loop
		local date = AIDate.GetCurrentDate();
		local currentYear = AIDate.GetYear(date);
		local lastYear = currentYear;
		local endYear = currentYear + 10;	// last year to chance search radius for new town connections
		
		while(true){
			if(currentYear != lastYear && currentYear < endYear) {
				lastYear = currentYear;
				MAX_DISTANCE += 10;
				MIN_DISTANCE -= 1;
			}
			
			if(AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > expandMoneyMin) {
				if(lastMail1 != -1 && lastMail2 != -1) {
					// add buses and mail trucks where there is plenty cargo
					local lorry1 = AIStation.GetStationID(lastMail1);
					local lorry2 = AIStation.GetStationID(lastMail2);
					if(min(AIStation.GetCargoWaiting(lorry1, mailCargoId), AIStation.GetCargoWaiting(lorry2, mailCargoId)) > 3 * AIVehicle.GetCapacity(lastMailVehicle, mailCargoId)) {
						AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						Debug("Adding mailtrucks at " + AIStation.GetName(lorry1) + " to " + AIStation.GetName(lorry2));
						connectMail(lastMail1, lastMail2);
					}
				}
				if(lastTown1 != -1 && lastTown2 != -1) {
					// add buses and mail trucks where there is plenty cargo
					local pass1 = AIStation.GetStationID(lastTown1);
					local pass2 = AIStation.GetStationID(lastTown2);
					if(min(AIStation.GetCargoWaiting(pass1, passengerCargoId), AIStation.GetCargoWaiting(pass2, passengerCargoId)) > 3 * AIVehicle.GetCapacity(lastPassengerVehicle, passengerCargoId) && lastMaxBusses) {
						AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						Debug("Adding busses at " + AIStation.GetName(pass1) + " to " + AIStation.GetName(pass2));
						if(connectStations(lastTown1, lastTown2))
							lastMaxBusses--;
					}
				}
			}
			
			if(AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 80000 && passengerEngine != -1) {
				Debug("We've got enough money to connect 2 more towns!");
				if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) < 125000)
					AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
				Debug("sub count: " + AISubsidyList().Count() + "; building one: " + subsidyBuilt);
				if(AISubsidyList().Count() && !subsidyBuilt) {
					connectSubsidy(AISubsidyList().Begin());
					subsidyBuilt = true;
				} else {
					connect2Towns();
				}
			}
						
			manageLoan();
			checkEvents();
			
			this.Sleep(1);
			
			date = AIDate.GetCurrentDate();
			currentYear = AIDate.GetYear(date);
		}
	}

	function getEngine(cargoID) {
		local engine_list = AIEngineList(AIVehicle.VT_ROAD);
		engine_list.Valuate(AIEngine.CanRefitCargo, cargoID);
		engine_list.KeepValue(1);

		engine_list.Valuate(AIEngine.GetRoadType);
		engine_list.KeepValue(AIRoad.ROADTYPE_ROAD);

		engine_list.Valuate(AIEngine.GetCapacity)
		engine_list.KeepTop(1);
		if (engine_list.Count() == 0) {
			Warning("No suitable " + AICargo.GetCargoLabel(cargoID) + " engines available (yet?)!!");
			return -1;
		}
		
		local engine = engine_list.Begin();
		Debug("Picked engine for cargo " + AICargo.GetCargoLabel(cargoID) + ": " + AIEngine.GetName(engine));

		return engine;
	}
	
	// 6.3 tiles per day for passengers? // TODO: check or calculate
	function incomeSort(targetTown, startTile, cargoID, engID) {
		local distance = AIMap.DistanceManhattan(startTile, AITown.GetLocation(targetTown));
		local days = distance/7 + 1;	//underestimate
		//Debug(AITown.GetName(targetTown) + " at distance: " + distance + ", days: " + days + " income: " + AICargo.GetCargoIncome(cargoID, distance, days) + " maxProd: " + AITown.GetMaxProduction(targetTown, cargoID));
		return AICargo.GetCargoIncome(cargoID, distance, days) * (AITown.GetMaxProduction(targetTown, cargoID) - AITown.GetLastMonthTransported(targetTown, cargoID)) * (365/days);
	}
	
	function checkEvents() {
		while (AIEventController.IsEventWaiting()) {
			local e = AIEventController.GetNextEvent();
			//Debug("Got an eventA: " + e);
			switch (e.GetEventType()) {
				case AIEvent.AI_ET_VEHICLE_CRASHED:
					local ec = AIEventVehicleCrashed.Convert(e);
					local v  = ec.GetVehicleID();
					Debug("We have a crashed vehicle (" + v + ")");
					// TODO: Handle the crashed vehicle
					break;
				case AIEvent.AI_ET_ENGINE_AVAILABLE:
					Debug("New engine available; rechecking current engines...");
					passengerEngine = getEngine(passengerCargoId);
					expandMoneyMin = 2 * AIEngine.GetPrice(passengerEngine);

					mailEngine = getEngine(mailCargoId);
					expandMoneyMin = (2 * AIEngine.GetPrice(mailEngine)) < expandMoneyMin ? expandMoneyMin : 2 * AIEngine.GetPrice(mailEngine);
					break;
				case AIEvent.AI_ET_VEHICLE_LOST:
					Debug("A vehicle got lost!!!");
					// TODO: nuke vehicle??
					break;
				case AIEvent.AI_ET_VEHICLE_WAITING_IN_DEPOT:
					Debug("A vehicle arrived in the depot for selling");
					local ec = AIEventVehicleWaitingInDepot.Convert(e);
					local v  = ec.GetVehicleID();
					AIVehicle.SellVehicle(v);
					break;
				case AIEvent. AI_ET_VEHICLE_UNPROFITABLE:
					Debug("Unprofitable vehicle detected; sending to depot");
					// TODO: evaluate on one more year?
					local ec = AIEventVehicleUnprofitable.Convert(e);
					local v  = ec.GetVehicleID();
					AIVehicle.SendVehicleToDepot(v);
					break;
				case AIEvent.AI_ET_SUBSIDY_OFFER:
					local ec = AIEventSubsidyOffer.Convert(e);
					local s = ec.GetSubsidyID();
					Debug("New subsidy detected: " + s);
					if (AICompany.GetMaxLoanAmount() - AICompany.GetLoanAmount() + AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 70000) {
						Debug("We've got enough money to get the subsidized route!");
						if (AICompany.GetBankBalance(AICompany.COMPANY_SELF) < 125000)
							AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
						connectSubsidy(s);
					} else {
						Warning("Not enough money to get this one now.");
					}
					break;
				case AIEvent.AI_ET_SUBSIDY_AWARDED:
					subsidyBuilt = false;
					local ec = AIEventSubsidyAwarded.Convert(e);
					local s = ec.GetSubsidyID();
					if (AICompany.IsMine(AISubsidy.GetAwardedTo(s)))
						Debug("We got the subsidy!!!!");
					else
						Warning("!!! We failed to win the subsidy!!!!");
					break;
				case AIEvent.AI_ET_SUBSIDY_OFFER_EXPIRED:
					subsidyBuilt = false;
					Debug("subsidy expired");
					break;
				default:
					//Debug("Unhandled event:" + e);
			}
		}
	}
	
	function manageLoan() {
		//repay loan
		local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
		local balanceMod = (balance / AICompany.GetLoanInterval()) * AICompany.GetLoanInterval();
		local loan = AICompany.GetLoanAmount();
		
		if (balance < 0) {
			// not good... we lost money?
			AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
			balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
			balanceMod = (balance / AICompany.GetLoanInterval()) * AICompany.GetLoanInterval();
			loan = AICompany.GetLoanAmount();
		}
		
		if (loan > 0 && balanceMod > 0) {
			if (balanceMod > 100000)
				balanceMod = 100000;
			AICompany.SetLoanAmount(loan - balanceMod);
		}
	}

	/**
	 * 2 passenger stations build, make road and then place busses
	 */
	function connectTownsByRoad(starts, goals) {
		pathfinder.cost.max_cost = 15 * AIMap.DistanceManhattan(starts[0], goals[0]);
		pathfinder.InitializePath(starts, goals);
		local path = false;
		local i = 0;
		local err;
		//local iterations = 300 * AIMap.DistanceManhattan(starts[0], goals[0]); // if we pass this; something is wrong
		local iterations = -1;	// TODO limit to reasonable amount, elevate if fails
		local startTick = GetTick();
		// TODO check whether higher values helps finding faster
		Debug("pathfinding for max " + pathfinder.cost.max_cost + " cost and iterations: " + iterations);
		path = pathfinder.FindPath(iterations);
		Debug("Total ticks passed: " + (GetTick() - startTick));

		if(!path) {
			if (path == false) {
				Error("!! No path found, too little iterations...");	// impossible atm
				pathfinder.reset();
				return false;
			} else {
				Warning("!! No path found, too expensive, retrying...");
				pathfinder.reset();
				pathfinder.cost.max_cost = 30 * AIMap.DistanceManhattan(starts[0], goals[0]);
				pathfinder.InitializePath(starts, goals);
				path = pathfinder.FindPath(iterations);
			}
		}
		if(!path) {
			pathfinder.reset();
			Error("Impossible to create route, way too expensive");
			return false;
		}
		Debug("path found for cost: " + path.GetCost() + ", building");

		while (path != null) {
			local par = path.GetParent();
			if (par != null) {
				local last_node = path.GetTile();
				if (AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) == 1 ) {
					if (!AIRoad.BuildRoad(path.GetTile(), par.GetTile())) {
						switch(AIError.GetLastError()) {
							case AIError.ERR_VEHICLE_IN_THE_WAY:
								Warning("Building road failed temporary: vehicle in the way...");
								// try again till the stupid vehicle moved?
								while(!AIRoad.BuildRoad(path.GetTile(), par.GetTile()));
								break;
							case AIRoad.ERR_ROAD_WORKS_IN_PROGRESS:
								Warning("Building road failed: road works in progress");
								// good, someone else is building this, skip tile
								break;
							case AIError.ERR_ALREADY_BUILT:
								// Debug("Building road failed: already built");
								// even better; someone else already built this; silent ignore
								break;
							case AIError.ERR_LAND_SLOPED_WRONG:
								Error("Building road failed: wrong land slope");
								// not good; redo from start or terraform?
								return connectTownsByRoad(starts, goals);
							case AIError.ERR_AREA_NOT_CLEAR:
								Warning("Building road failed: not clear, demolishing tile...");
								AISign.BuildSign(path.GetTile(), "I'm on your land, demolishing your stuffs");
								// demolish and retry?
								if (AITile.DemolishTile(path.GetTile())) {
									AIRoad.BuildRoad(path.GetTile(), par.GetTile());
								} else {	// happens when other ai builds after calc
									return connectTownsByRoad(starts, goals);
								}
								break;
							case AIRoad.ERR_ROAD_ONE_WAY_ROADS_CANNOT_HAVE_JUNCTIONS:
								Error("Building road failed: no junctions on one way roads allowed");
								// can't happen? Not building those
								return false;
							case AIError.ERR_NOT_ENOUGH_CASH:
								Error("Panic: Out of cash!!");
								return false;
							default:
								Warning("Unknown error during road building " + AIError.GetLastErrorString());
								break;	// no clue what this is; silent ignore for now
						}
					}
				} else {
					/* Build a bridge or tunnel. */
					if (!AIBridge.IsBridgeTile(path.GetTile()) && !AITunnel.IsTunnelTile(path.GetTile())) {
						/* If it was a road tile, demolish it first. Do this to work around expended roadbits. */
						if (AIRoad.IsRoadTile(path.GetTile())) AITile.DemolishTile(path.GetTile());
						if (AITunnel.GetOtherTunnelEnd(path.GetTile()) == par.GetTile()) {
							if (!AITunnel.BuildTunnel(AIVehicle.VT_ROAD, path.GetTile())) {
								/* An error occured while building a tunnel. TODO: handle it. */
								// TODO expected to fail once per tunnel atm
								// AISign.BuildSign(path.GetTile(), ("Tunnel fail"));
								Error("Couldn't build tunnel (normal once per tunnel atm");
							}
						} else {
							local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) + 1);
							bridge_list.Valuate(AIBridge.GetMaxSpeed);
							bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
							if (!AIBridge.BuildBridge(AIVehicle.VT_ROAD, bridge_list.Begin(), path.GetTile(), par.GetTile())) {
								/* An error occured while building a bridge. TODO: handle it. */
								// TODO expected to fail once per bridge atm
								// AISign.BuildSign(path.GetTile(), ("Bridge fail"));
								Error("Couldn't build bridge (normal once per bridge atm)");
							}
						}
					}
				}
			}
			path = par;
		}

		/*while (path != null) {
			if (path.GetParent() != null) {
				local par = path.GetParent().GetTile();
				if (!AIRoad.AreRoadTilesConnected(path.GetTile(), par) && (!AIBridge.IsBridgeTile(path.GetTile()) || AIBridge.GetOtherBridgeEnd(path.GetTile()) != par)) {
					AIRoad.BuildRoad(path.GetTile(), par);
				}
			}
			path = path.GetParent();
		}*/
		Debug("path done");
		return true;
	}

	function connectStations(station1, station2) {		
		local bus1 = AIVehicle.BuildVehicle(myStationsTileList.GetValue(station1), passengerEngine);
		if (!AIVehicle.IsValidVehicle(bus1))
			return false;
		if(!AIVehicle.RefitVehicle(bus1, passengerCargoId)) {
			Error("Error refitting bus: " + AIError.GetLastErrorString());
			return false;
		}

		Debug("Built bus: " + bus1);
		lastPassengerVehicle = bus1;
		AIOrder.AppendOrder(bus1, station1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(bus1, station2, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIVehicle.StartStopVehicle(bus1);
		local bus2 = AIVehicle.BuildVehicle(myStationsTileList.GetValue(station2), passengerEngine);
		if (!AIVehicle.IsValidVehicle(bus2))
			return false; 	// TODO change to return 1 for 1 success
		if(!AIVehicle.RefitVehicle(bus2, passengerCargoId)) {
			Error("Error refitting bus: " + AIError.GetLastErrorString());
			return false; // TODO
		}
		Debug("Built bus: " + bus2);
		AIOrder.AppendOrder(bus2, station2, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(bus2, station1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIVehicle.StartStopVehicle(bus2);
		return true;	// TODO change to 2
	}
	
	function connectMail(lorry1, lorry2) {
		local mail1 = AIVehicle.BuildVehicle(myLorryTileList.GetValue(lorry1), mailEngine);
		if(!AIVehicle.IsValidVehicle(mail1))
			return false;
		Debug("Built a mail truck: " + mail1 + " at location " + lorry1);
		lastMailVehicle = mail1;
		AIOrder.AppendOrder(mail1, lorry1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(mail1, lorry2, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIVehicle.StartStopVehicle(mail1);
		local mail2 = AIVehicle.BuildVehicle(myLorryTileList.GetValue(lorry2), mailEngine);
		if(!AIVehicle.IsValidVehicle(mail2))
			return false;
		Debug("Built a mail truck: " + mail2 + " at location " + lorry2);
		AIOrder.AppendOrder(mail2, lorry2, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIOrder.AppendOrder(mail2, lorry1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
		AIVehicle.StartStopVehicle(mail2);
	}
	
	function getRoadTile(t) {
		local adjacent = AITileList();
		adjacent.AddTile(t - AIMap.GetTileIndex(1,0));
		adjacent.AddTile(t - AIMap.GetTileIndex(0,1));
		adjacent.AddTile(t - AIMap.GetTileIndex(-1,0));
		adjacent.AddTile(t - AIMap.GetTileIndex(0,-1));
		adjacent.Valuate(AIRoad.IsRoadTile);
		adjacent.KeepValue(1);
		adjacent.Valuate(AIRoad.IsRoadDepotTile);
		adjacent.KeepValue(0);
		adjacent.Valuate(AIRoad.IsRoadStationTile);
		adjacent.KeepValue(0);
		if (adjacent.Count())
			return adjacent.Begin();
		else
			return null;
	}

	/**
	 * hack to support basic sqrt function for other functions
	 * TODO: consider using lookup table for speed reasons (as suggested by Truebrain)
	 */
	function sqrt(i) {
		if (i == 0)
			return 0;   // Avoid divide by zero
		local n = (i / 2) + 1;       // Initial estimate, never low
		local n1 = (n + (i / n)) / 2;
		while (n1 < n) {
			n = n1;
			n1 = (n + (i / n)) / 2;
		}
		return n;
	}

	function buildNonInlinePassengerStation(town) {
		Debug("Building non inline station...");
		// Find empty square as close to town centre as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 4; // search larger area
		local area = AITileList();

		while (curRange < maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(curRange, curRange), AITown.GetLocation(town) + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AITile.IsBuildable);
			area.KeepValue(1);
			if (area.Count()) {
				//Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = getRoadTile(t);
					if(opening) {
						// Debug("tile " + t + " has opening " + opening);
						if (AIRoad.BuildRoad(opening, t) && AIRoad.BuildRoadStation(t, opening, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT)) {
							//AISign.BuildSign(t, ("bus " + AITown.GetName(town)));
							return t;
						}
						Error("Error building passenger station " + AIError.GetLastErrorString());
						//AISign.BuildSign(t, "" + AIRoad.GetNeighbourRoadCount(t));
						if(AIError.GetLastError() == AIRoad.ERR_ROAD_CANNOT_BUILD_ON_TOWN_ROAD)
							return buildNonInlinePassengerStation(town);

					} //else {
						// Debug("tile" + t + " has no opening");
					//}
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible station location found");
		return null;
	}
	
	function buildPassengerStation(town) {
		Debug("Building station...");
		// Find empty square as close to town centre as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 2; //TODO check value correctness 
		local area = AITileList();
		
		while (curRange < maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(curRange, curRange), AITown.GetLocation(town) + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AIRoad.IsRoadTile);
			area.KeepValue(1);
			area.Valuate(AIRoad.IsDriveThroughRoadStationTile);
			area.KeepValue(0);
			area.Valuate(AIRoad.GetNeighbourRoadCount);
			area.KeepValue(2);	// entrance and exit; allow 1 as well?
			if (area.Count()) {
				//Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = getRoadTile(t);
					if(opening) {
						// Debug("tile " + t + " has opening " + opening);
						if (/*AIRoad.BuildRoad(opening, t) &&*/ AIRoad.BuildDriveThroughRoadStation(t, opening, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT)) {
							//AISign.BuildSign(t, ("bus " + AITown.GetName(town)));
							return t;
						}
						if(AIError.GetLastError() == AIRoad.ERR_ROAD_CANNOT_BUILD_ON_TOWN_ROAD)
							return buildNonInlinePassengerStation(town);
						Error("Error building passenger station " + AIError.GetLastErrorString());
						//AISign.BuildSign(t, AIError.GetLastErrorString());

					} //else {
						// Debug("tile" + t + " has no opening");
					//}
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible station location found");
		return null;
	}
	
	function buildLorryStation(town) {	// TODO check mail acceptance
		if (mailEngine == -1) {
			Warning("No mail engine available, not building mail stations");
			return null;
		}

		Debug("Building lorry station...");
		// Find empty square as close to town centre as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 1; //TODO check value correctness
		local area = AITileList();
		
		while (curRange < maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(curRange, curRange), AITown.GetLocation(town) + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AITile.IsBuildable);
			area.KeepValue(1);
			if (area.Count()) {
				//Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = getRoadTile(t);
					if(opening) {
						// Debug("tile " + t + " has opening " + opening);
						if (AIRoad.BuildRoad(opening, t) && AIRoad.BuildDriveThroughRoadStation(t, opening, AIRoad.ROADVEHTYPE_TRUCK, AIStation.STATION_JOIN_ADJACENT)) {
							//AISign.BuildSign(t, ("Mail " + AITown.GetName(town)));
							return t;
						}
						//Debug("tile " + t + " has wrong slope!!!");
					} //else {
						// Debug("tile" + t + " has no opening");
					//}
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible mail location found");
		return null;
	}
	
	function buildHQ(town) {
		Debug("building HQ...");
		// Find empty 2x2 square as close to town centre as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 5; //TODO check value correctness
		local area = AITileList();
		
		while (curRange < maxRange) {
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(curRange, curRange), AITown.GetLocation(town) + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AITile.IsBuildableRectangle, 2, 2);
			area.KeepValue(1);
			if (area.Count()) {
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = getRoadTile(t);
					if(opening) {
						if (AICompany.BuildCompanyHQ(t)) {
							AISign.BuildSign(t, ("HQ OTVI"));
							return t;
						}
					}
				}
				curRange++;
			} else {
				curRange++;
			}
		}
		Debug("No possible HQ location found");
		return null;
	}
	
	function buildDepot(station, town) {
		Debug("Building depot...");
		// Find empty square as close to station as possible
		local spotFound = false;
		local curRange = 1;
		local maxRange = sqrt(AITown.GetPopulation(town)/100) + 4; //TODO check value correctness
		local area = AITileList();
		
		while (curRange < maxRange) {
			//Debug("looking with range " + curRange);
			area.AddRectangle(station - AIMap.GetTileIndex(curRange, curRange), station + AIMap.GetTileIndex(curRange, curRange));
			area.Valuate(AITile.IsBuildable);
			area.KeepValue(1);
			if (area.Count()) {
				//Debug("Found " + area.Count() + " possible options");
				for (local t = area.Begin(); area.HasNext(); t = area.Next()) {
					local opening = getRoadTile(t);
					if(opening) {
						//Debug("tile " + t + " has opening " + opening);
						if (AIRoad.BuildRoad(t, opening) && AIRoad.BuildRoadDepot(t, opening)) {
							//AISign.BuildSign(t, ("depot"));
							return t;
						}
						//Debug("tile " + t + " has wrong slope!!!");
					} //else {
						//Debug("tile" + t + " has no opening");
					//}
				}
				curRange++; // the found options had no road connections; enlarge search area
			} else {
				curRange++;
			}
		}
		Debug("No possible depot location found");
		return null;
	}
	
	function connect2Industries(subID) {
		local station1 = buildLorryStation(AIIndustry.GetLocation(AISubsidy.GetSource(subID)));
		if (!station1)
			return false; // bugger
		local depot1 = buildDepot(station1);
		myStationsTileList.AddItem(station1, depot1);
		local loc2;
		if (AISubsidy.DestinationIsTown(subID))
			loc2 = AITown.GetLocation(AISubsidy.GetDestination(subID));
		else
			loc2 = AIIndustry.GetLocation(AISubsidy.GetDestination(subID));
		local station2 = buildLorryStation(loc2);
		if (!station2)
			return false; // bugger
		local depot2 = buildDepot(station2);
		myStationsTileList.AddItem(station2, depot2);

		if(!depot1 && !depot2)
			return false;

		AILog.Info("Route between 2 industries is " + AITile.GetDistanceManhattanToTile(station1, station2));
		if(!connectTownsByRoad(station1, station2))
			connectTownsByRoad(station1, station2);	// Try once more if it failed
		//connectTownsByRoad(station1, depot1);
		//connectTownsByRoad(station2, depot2);
		connectIndustries(station1, station2, AISubsidy.GetCargoType(subID));
	}

	function connectIndustries(station1, station2, cargoID) {	
		local engine = getEngine(cargoID);
		local count = (AITile.GetDistanceManhattanToTile(station1, station2) / 10) + 1;
		for (local i = 0; i < count; i++) {
			local vehicle = AIVehicle.BuildVehicle(myStationsTileList.GetValue(station1), engine);
			Debug("Built a vehicle " + vehicle + " which is valid: " + AIVehicle.IsValidVehicle(vehicle));
			AIOrder.AppendOrder(vehicle, station1, AIOrder.AIOF_FULL_LOAD | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
			AIOrder.AppendOrder(vehicle, station2, AIOrder.AIOF_NONE | AIOrder.AIOF_NON_STOP_INTERMEDIATE);
			AIVehicle.StartStopVehicle(vehicle);
			if (!AIVehicle.IsValidVehicle(vehicle))
				return false;
		}
		return true;
	}

	function connectTowns(subID) {
		biggestTown = AISubsidy.GetSource(subID);
		
		Debug("First subsidy town: " + AITown.GetName(biggestTown));
		local station1 = buildPassengerStation(biggestTown);
		if(!station1) {
			Error("!!! Building station failed !!!");
			return false;
		}

		local mail1;

		local depot1 = myTownList.GetValue(biggestTown);
		if (!depot1) {
			mail1 = buildLorryStation(biggestTown);
			depot1 = buildDepot(station1, biggestTown);
			if(!depot1) {
				// to prevent it from choosing this "bad" town again
				myTownList.SetValue(biggestTown, -1);
				return false;
			}
			myTownList.SetValue(biggestTown, depot1);
			if (mail1) { 
				myLorryTileList.AddItem(mail1, depot1);
				myDepotsTileList.AddItem(depot1, mail1);
			}
		} else {
			mail1 = myDepotsTileList.GetValue(myTownList.GetValue(biggestTown));
			if (!mail1) {
				mail1 = buildLorryStation(biggestTown);
				if (mail1) { 
					myLorryTileList.AddItem(mail1, depot1);
					myDepotsTileList.AddItem(depot1, mail1);
				}
			}
		}
		myStationsTileList.AddItem(station1, depot1);
		
		local connectTown = AISubsidy.GetDestination(subID);
		
		Debug("Connecting subsidy town: " + AITown.GetName(connectTown) + " at a manhattan distance of " + AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown)));

		local station2 = buildPassengerStation(connectTown);
		if(!station2) {
			Error("!!! Building station failed !!!");
			return false;
		}

		local mail2;

		local depot2 = myTownList.GetValue(connectTown);
		if (!depot2) {
			if(mail1)
				mail2 = buildLorryStation(connectTown);
			depot2 = buildDepot(station2, connectTown);
			if(!depot2)
				return false;
			myTownList.SetValue(connectTown, depot2);
			if (mail2) { 
				myLorryTileList.AddItem(mail2, depot2);
				myDepotsTileList.AddItem(depot2, mail2);
			}
		} else {
			mail2 = myDepotsTileList.GetValue(myTownList.GetValue(connectTown));
			if (!mail2 && mail1) {
				if(mail1)
					mail2 = buildLorryStation(connectTown);
				if (mail2) { 
					myLorryTileList.AddItem(mail2, depot2);
					myDepotsTileList.AddItem(depot2, mail2);
				}
			}
		}
		myStationsTileList.AddItem(station2, depot2);
			
		local distance = AITile.GetDistanceManhattanToTile(station1, station2);
		Debug("Route between 2 cities is " + distance);
		local opening1 = AIRoad.GetRoadDepotFrontTile(depot1);
		local opening2 = AIRoad.GetRoadDepotFrontTile(depot2);
		local starts = mail1 ? [opening1, mail1, station1] : [opening1, station1];
		local goals = mail2 ? [opening2, mail2, station2] : [opening2, station2];
		if(!connectTownsByRoad(starts, goals))
			connectTownsByRoad(starts, goals);	// Try once more if it failed
		if (mail1 && mail2) {
			// make sure the mail stations are connected to the road build for the passengers
			//connectTownsByRoad(station1, mail1);
			//connectTownsByRoad(station2, mail2);
			// sending out mail truck first, otherwise it might get stopped at a drive through station by our own bus
			lastMail1 = mail1;
			lastMail2 = mail2;
			connectMail(mail1, mail2);
		}
		
		//we picked BIG towns, lets try to build more busses in 1 go
		local pop1 = AITown.GetPopulation(biggestTown);
		local pop2 = AITown.GetPopulation(connectTown);
		local engine = getEngine(passengerCargoId);
		// TODO: see whether this value max sense:
		local divider = AIEngine.GetCapacity(engine) * 5;
		local maxBussesPop = (pop1 < pop2 ? pop1 : pop2) / divider;	// built a bus for every hundred passengers in the smallest town
		local maxBussesDist = (distance / (AIEngine.GetMaxSpeed(engine) / 10));	// tweaked to 10. May need to be raised to 12...
		local maxBusses = maxBussesPop < maxBussesDist ? maxBussesPop : maxBussesDist;
		maxBusses++;	// plus 2 busses for both stops
		Debug("Want to deploy " + (maxBusses * 2) + " busses for this route (pop/dist:) " + ((maxBussesPop + 1) * 2) + "/" + ((maxBussesDist + 1) * 2));
		while(maxBusses-- && AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 12500)	// actually 12.500 is only enough for 1 bus
			connectStations(station1, station2);
		if(maxBusses) {
			lastMaxBusses = maxBusses;
			lastTown1 = station1;
			lastTown2 = station2;
		}
	}

	function connect2Towns() {
		// look for biggest towns
		local townlist = AITownList();
		townlist.Valuate(AITown.GetMaxProduction, mailCargoId);
		townlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
		biggestTown = townlist.Begin();
		
		// skip towns already taken
		while (myTownList.GetValue(biggestTown)) {
			if(!townlist.HasNext()) {
				Debug("No more towns left");
				manageLoan();
				Sleep(100);
				return;
			}
			//Debug("Skipping town (already taken): " + AITown.GetName(biggestTown));
			biggestTown = townlist.Next();
		}
		
		Debug("biggest town: " + AITown.GetName(biggestTown));
		local station1 = buildPassengerStation(biggestTown);
		local mail1;
		while (!station1) {
			Debug("!!! Building station failed !!!");
			biggestTown = townlist.Next();
			while (myTownList.GetValue(biggestTown))
				biggestTown = townlist.Next();
			station1 = buildPassengerStation(biggestTown);
		}
		local depot1 = myTownList.GetValue(biggestTown);
		if (station1) {
			if (!depot1) {
				mail1 = buildLorryStation(biggestTown);
				depot1 = buildDepot(station1, biggestTown);
				if(!depot1) {
					// to prevent it from choosing this "bad" town again
					myTownList.SetValue(biggestTown, -1);
					return false;
				}
				myTownList.SetValue(biggestTown, depot1);
				if (mail1) { 
					myLorryTileList.AddItem(mail1, depot1);
					myDepotsTileList.AddItem(depot1, mail1);
				}
			} else {
				mail1 = myDepotsTileList.GetValue(myTownList.GetValue(biggestTown));
				if (!mail1) {
					mail1 = buildLorryStation(biggestTown);
					if (mail1) { 
						myLorryTileList.AddItem(mail1, depot1);
						myDepotsTileList.AddItem(depot1, mail1);
					}
				}
			}
			myStationsTileList.AddItem(station1, depot1);
		} else {
			return false;
		}
		
		// find connecting town between MIN and MAX manhattan distance
		townlist = AITownList();
		townlist.Valuate(incomeSort, station1, mailCargoId, mailEngine);
		townlist.Sort(AIAbstractList.SORT_BY_VALUE, false);
		local connectTown = townlist.Begin();
		local distance = AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown));
		while (distance > MAX_DISTANCE || distance < MIN_DISTANCE) {
			Debug("considering town: " + AITown.GetName(connectTown) + " at distance " + AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown)) + " value: " + townlist.GetValue(connectTown));
			if (!townlist.HasNext()) {	// TODO: add check like: OR skipped 20 or value < maxvalue/4 or something
				Debug("Out of towns");
				return; // too bad, out of towns
			}
			connectTown = townlist.Next();
			distance = AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown));
		}
		
		Debug("Connecting town: " + AITown.GetName(connectTown) + " at a manhattan distance of " + AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown)) + " value: " + townlist.GetValue(connectTown));
		local station2 = buildPassengerStation(connectTown);
		while (!station2) {
			Debug("!!! Building station failed !!!");
			connectTown = townlist.Next();
			distance = AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown));
			while (distance > MAX_DISTANCE || distance < MIN_DISTANCE) {
				Debug("considering town: " + AITown.GetName(connectTown) + " at distance " + AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown)) + " value: " + townlist.GetValue(connectTown));
				if (!townlist.HasNext()) {	// TODO: add check like: OR skipped 20 or value < maxvalue/4 or something
					Debug("Out of towns");
					return; // too bad, out of towns
				}
				connectTown = townlist.Next();
				distance = AITile.GetDistanceManhattanToTile(AITown.GetLocation(biggestTown), AITown.GetLocation(connectTown));
			}
			station2 = buildPassengerStation(connectTown);
		}

		local mail2;

		local depot2 = myTownList.GetValue(connectTown);
		if (station2) {
			if (!depot2) {
				if(mail1)
					mail2 = buildLorryStation(connectTown);
				depot2 = buildDepot(station2, connectTown);
				if(!depot2)
					return false;
				myTownList.SetValue(connectTown, depot2);
				if (mail2) { 
					myLorryTileList.AddItem(mail2, depot2);
					myDepotsTileList.AddItem(depot2, mail2);
				}
			} else {
				mail2 = myDepotsTileList.GetValue(myTownList.GetValue(connectTown));
				if (!mail2 && mail1) {
					if(mail1)
						mail2 = buildLorryStation(connectTown);
					if (mail2) { 
						myLorryTileList.AddItem(mail2, depot2);
						myDepotsTileList.AddItem(depot2, mail2);
					}
				}
			}
			myStationsTileList.AddItem(station2, depot2);
		}
			
		Debug("Route between 2 cities is " + AITile.GetDistanceManhattanToTile(station1, station2));
		local opening1 = AIRoad.GetRoadDepotFrontTile(depot1);
		local opening2 = AIRoad.GetRoadDepotFrontTile(depot2);
		local starts = mail1 ? [opening1, mail1, station1] : [opening1, station1];
		local goals = mail2 ? [opening2, mail2, station2] : [opening2, station2];
		if(!connectTownsByRoad(starts, goals))
			connectTownsByRoad(starts, goals);	// Try once more if it failed
		if (mail1 && mail2) {
			// make sure the mail stations are connected to the road build for the passengers
			//connectTownsByRoad(station1, mail1);
			//connectTownsByRoad(station2, mail2);
			// sending out mail truck first, otherwise it might get stopped at a drive through station by our own bus
			lastMail1 = mail1;
			lastMail2 = mail2;
			connectMail(mail1, mail2);
		}
		
		//we picked BIG towns, lets try to build more busses in 1 go
		local pop1 = AITown.GetPopulation(biggestTown);
		local pop2 = AITown.GetPopulation(connectTown);
		local engine = getEngine(passengerCargoId);
		// TODO: see whether this value max sense:
		local divider = AIEngine.GetCapacity(engine) * 5;
		local maxBussesPop = (pop1 < pop2 ? pop1 : pop2) / divider;	// built a bus for every hundred passengers in the smallest town
		local maxBussesDist = (distance / (AIEngine.GetMaxSpeed(engine) / 10));	// tweaked to 10. May need to be raised to 12...
		local maxBusses = maxBussesPop < maxBussesDist ? maxBussesPop : maxBussesDist;
		maxBusses++;	// plus 2 busses for both stops
		Debug("Want to deploy " + (maxBusses * 2) + " busses for this route (pop/dist:) " + ((maxBussesPop + 1) * 2) + "/" + ((maxBussesDist + 1) * 2));
		while(maxBusses-- && AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 12500)	// actually 12.500 is only enough for 1 bus
			connectStations(station1, station2);
		if(maxBusses) {
			lastMaxBusses = maxBusses;
			lastTown1 = station1;
			lastTown2 = station2;
		}
	}

	function connectSubsidy(subID) {		
		if (AISubsidy.SourceIsTown(subID)) {
			connectTowns(subID);
		} else {
			connect2Industries(subID);
		}
	}
}

